local types = {};

luanet.load_assembly("log4net");

types["log4net.LogManager"] = luanet.import_type("log4net.LogManager");

local rootLogger = "AtlasSystems.Addons.ReprintsDeskDSP"
local log = types["log4net.LogManager"].GetLogger(rootLogger);

local Settings = {};

local reprintsDesk = {};

reprintsDesk.Order = {};
reprintsDesk.Order.Controls = nil;

local INCLUSIVE_PAGES_DEFAULT = 8;
local ORDER_TYPE_DEFAULT = 4;
local DELIVERY_METHOD_DEFAULT = 5;
local PRICE_TYPE_DEFAULT = 2;
local PRICE_TYPE_ALTERNATE = 1;
local ORDER_USER_DEFAULT = "*";

function Init()
	log:Info("Initializing Reprints Desk Unmediated Addon");	
	InitializeSettings();	
	AddLocalInfoTagReplacements();
	
	RegisterSystemEventHandler("GenerateFulfillmentData","RetrieveReprintsDeskEstimate");
    RegisterSystemEventHandler("AutomationDecision","PlaceReprintsDeskOrder");
end

function InitializeSettings()
	log:Debug("ReprintsDeskUnmediated [InitializeSettings] BEGIN");
	-- Config Settings
	Settings.UseReprintsDeskProductionServer = GetSetting("UseReprintsDeskProductionServer");
	Settings.Username = GetSetting("ReprintsDeskAPIUsername");
	Settings.Password = GetSetting("ReprintsDeskAPIPassword");
	Settings.ChargesField = GetSetting("ChargesField");
	Settings.OrderPlacedQueue = GetSetting("OrderPlacedQueue");
	Settings.ReprintsDeskOrderIdField = GetSetting("ReprintsDeskOrderIdField");
	Settings.ReprintsDeskLenderSymbol = GetSetting("ReprintsDeskLenderSymbol");
	Settings.ErrorQueue = GetSetting("FailureQueue");
	Settings.NotePrefix = GetSetting("NotePrefix");

	--Service Settings
	ReprintsDeskService.IsProduction =  Settings.UseReprintsDeskProductionServer;
	ReprintsDeskService.Username = Settings.Username;
	ReprintsDeskService.Password = Settings.Password;

	-- Service client
	ReprintsDeskService.UserAgent = "ILLiad/" .. ILLiad.Version;

	log:Debug("ReprintsDeskUnmediated [InitializeSettings] COMPLETE");
end

function AddLocalInfoTagReplacements()
	log:Debug("[AddLocalInfoTagReplacements]");

	local connection = CreateManagedDatabaseConnection();

    local retrievedLocalInfo, localInfoResults = pcall(function ()
        connection:Connect();
        connection.QueryString = "SELECT * FROM LocalInfo";
		Utility.Log(connection.QueryString);

        local queryResults = connection:Execute();
		local parsedResults = {};	
		log:DebugFormat("RowCount: {0}", queryResults.Rows.Count);

		for i = 0, queryResults.Rows.Count - 1 do
			local nvtgc = string.upper(queryResults.Rows:get_Item(i):get_Item("NVTGC"));
			parsedResults[nvtgc] = {};
			
			for j = 0, queryResults.Columns.Count - 1 do
			local columnName = queryResults.Columns:get_Item(j).ColumnName;
				parsedResults[nvtgc][columnName] = queryResults.Rows:get_Item(i):get_Item(columnName);
				log:DebugFormat("LocalInfo| {0} | {1} = {2}",nvtgc, columnName,  queryResults.Rows:get_Item(i):get_Item(columnName));				
			 end
		end
        return parsedResults;
    end);

    if (retrievedLocalInfo) then
		TagProcessor.AddCustom('LocalInfo', function(field) 
			local nvtgc = string.upper(GetFieldValue("User","NVTGC"));
			return localInfoResults[nvtgc][field];
        end);
	else
		log:Warn("Unable to retrieve LocalInfo tag replacements");
	end

    connection:Dispose();
end

-- DSP Stage 2
-- Retrieve a RD estimate and add the information to the relevant event args.
function RetrieveReprintsDeskEstimate(eventArgs)
	log:Debug("[RetrieveReprintsDeskEstimate]");

	local validStandardNumber, cleanedStandardNumber = ILLiad.IsValidISxN(GetFieldValue("Transaction", "ISSN"));
	local year = tonumber(GetFieldValue("Transaction", "PhotoJournalYear"));
	local totalPages = ILLiad.InclusivePageCount(GetFieldValue("Transaction", "PhotoJournalInclusivePages"), INCLUSIVE_PAGES_DEFAULT);

	if (validStandardNumber and year and totalPages) then
		local input = {}

		input.TransactionNumber = tonumber(GetFieldValue("Transaction", "TransactionNumber"));
		input.StandardNumber = cleanedStandardNumber;
		input.Year = year;
		input.TotalPages = totalPages;
		input.PriceType = PRICE_TYPE_DEFAULT;

		log:Debug("Requesting price estimate");
		local result = ReprintsDeskService.GetOrderEstimate(input);
		ProcessReprintsDeskEstimate(result, eventArgs);
	else
		log:Warn("Unable to request estimate with current ISSN, year, and number of pages");
		eventArgs.Handled = false;
		eventArgs.Available = false;
		return;
	end
end

function ProcessReprintsDeskEstimate(result, eventArgs)
	log:Debug("[ProcessReprintsDeskEstimate]");
	Utility.Log(result, true);

	if (result.Error ~= nil) then
		log:Debug("Error encountered when retrieving estimate");
		eventArgs.Handled = false;
		eventArgs.Available = false;
		return;
	else
		if (result.Estimate) then
			log:Debug("ReprintsDesk estimate received")

			if ((result.Estimate.CopyrightCharge or -1) > 0) then
				eventArgs.Handled = true;
				eventArgs.Available = true;
				eventArgs.Cost = result.Estimate.ServiceCharge + result.Estimate.CopyrightCharge;
				eventArgs.TurnaroundTime = 3;
				eventArgs.Format = "pdf";
				eventArgs.CustomData = nil;
				eventArgs.CopyrightIncluded = true;
				return;
			else
				if (result.Input.PriceType == PRICE_TYPE_DEFAULT) then
					log:Debug("Requesting alternate estimate");
					result.Input.PriceType = PRICE_TYPE_ALTERNATE;

					local alternateResult = ReprintsDeskService.GetOrderEstimate(result.Input);
					ProcessReprintsDeskEstimate(alternateResult, eventArgs);
					return;
				end
			end
		end
	end

	log:Debug("Requested article is not available");
	eventArgs.Handled = true;
	eventArgs.Available = false;
end


-- DSP Stage 4
-- Place a RD order.
function PlaceReprintsDeskOrder(eventArgs)
	log:Debug("[PlaceReprintsDeskOrder]");

	if (GetFieldValue('Transaction', 'LendingLibrary') == Settings.ReprintsDeskLenderSymbol) then
		ExecuteCommand("AddNote", { GetFieldValue("Transaction", "TransactionNumber"), Settings.NotePrefix.."Reprints order already placed. Delete value in LendingLibrary if you wish to place another order."});
		eventArgs.Handled = true;
		return
	end

	order, validationErrors = CreateReprintsDeskOrder();

	if (#validationErrors > 0) then
		log:Warn("Validation errors when placing order");
		log:Warn(validationErrors);

		local errorMessage = "Validation errors found when preparing order:\r\n";

		for index, validationError in ipairs(validationErrors) do
			errorMessage = errorMessage .. "\r\n" .. validationError;
		end

		errorMessage = errorMessage .. "\r\n\r\nNote:Some fields can be changed in the OrderMap file.";

		log:DebugFormat("Unable to place ReprintsDesk order: {0}",errorMessage);
		eventArgs.Handled = false;
		return;
	end

	log:Debug("Placing order");
	local result =  ReprintsDeskService.PlaceOrder(order);

	if (result.Error) then
		log:Error(result.Error.Message);
	else
		if (result.Order and result.Order.OrderResult == 1) then
			-- Save fields to the transactions
			SetFieldValue('Transaction', Settings.ReprintsDeskOrderIdField, result.Order.OrderID);
			SetFieldValue('Transaction', 'LendingLibrary', Settings.ReprintsDeskLenderSymbol);
			SetFieldValue('Transaction', Settings.ChargesField, tonumber(eventArgs.FulfillmentData.Cost));
			SetFieldValue('Transaction', 'SystemID', 'OTH');
			SaveDataSource("Transaction");
			AddOrderNote(result.Input.TransactionNumber, result.Order);
			ExecuteCommand("Route", {GetFieldValue("Transaction", "TransactionNumber"), Settings.OrderPlacedQueue})

			eventArgs.Handled = true;
			return;
		end
	end
	
	log:Warn("Error placing order");
	local errorNote = "Unable to place ReprintsDesk order."
	ExecuteCommand("AddNote", {transactionNumber, errorNote});

	eventArgs.Handled = false;
	ExecuteCommand("Route", {GetFieldValue("Transaction", "TransactionNumber"), Settings.ErrorQueue})
end

function CreateReprintsDeskOrder()
	log:Debug("[CreateReprintsDeskOrder]");
	local validationErrors = {};
	local order = {};

	order.TransactionNumber = GetFieldValue("Transaction","TransactionNumber");

	-- Detail
	order.Detail = {};
	order.Detail.OrderTypeID = ORDER_TYPE_DEFAULT; -- Using default value per specification
	order.Detail.DeliveryMethodID = DELIVERY_METHOD_DEFAULT ; -- Using default value per specification

	for key, value in pairs(OrderMap.Detail) do
		if (value) then
			order.Detail[key] = TagProcessor.ReplaceTags(value) or "";
		else
			order.Detail[key] = "";
		end
	end

	-- User
	order.User = {};

	for key, value in pairs(OrderMap.User) do
		if (value) then
			order.User[key] = TagProcessor.ReplaceTags(value) or "";
		else
			order.User[key] = "";
		end
	end

	-- Delivery Profile
	order.DeliveryProfile = {};
	for key, value in pairs(OrderMap.DeliveryProfile) do
		if (value) then
			order.DeliveryProfile[key] = TagProcessor.ReplaceTags(value) or "";
		else
			order.DeliveryProfile[key] = "";
		end
	end

	-- Customer References
	order.CustomerReferences = {};

	for index, value in ipairs(OrderMap.CustomerReferences) do
		if (value) then
			order.CustomerReferences[index] = value or "";
		else
			order.CustomerReferences[index] = "";
		end
	end

	-- Customer Reference validation
	for index, value in ipairs(order.CustomerReferences) do
		ValidateField(order.CustomerReferences[index], "CustomerReferences[" .. index .. "]", 0, 1500, validationErrors);
	end

	-- User ensure values for set defaults
	order.User.FirstName = ValueOrDefault(order.User.FirstName, ORDER_USER_DEFAULT);
	order.User.LastName = ValueOrDefault(order.User.LastName, ORDER_USER_DEFAULT);
	order.User.BillingReference = ValueOrDefault(order.User.BillingReference, ORDER_USER_DEFAULT);

	-- User validation
	ValidateField(order.User.Username, "Username", 1, 64, validationErrors);
	ValidateField(order.User.Email, "Delivery E-mail", 1, 64, validationErrors);

	-- Delivery Profile
	order.DeliveryProfile.FirstName = ValueOrDefault(order.DeliveryProfile.FirstName,"*");
	order.DeliveryProfile.LastName = ValueOrDefault(order.DeliveryProfile.LastName, "*");
	order.DeliveryProfile.CompanyName = ValueOrDefault(order.DeliveryProfile.CompanyName,"*");
	order.DeliveryProfile.Address1 = ValueOrDefault(order.DeliveryProfile.Address,"*");
	order.DeliveryProfile.Address2 = ValueOrDefault(order.DeliveryProfile.AddressTwo, "*");
	order.DeliveryProfile.City = ValueOrDefault(order.DeliveryProfile.City, "*");
	order.DeliveryProfile.StateCode = ValueOrDefault(order.DeliveryProfile.StateCode, "**");
	order.DeliveryProfile.StateName = ValueOrDefault(order.DeliveryProfile.StateName, "*");
	order.DeliveryProfile.Zip = ValueOrDefault(order.DeliveryProfile.Zip, "*");
	order.DeliveryProfile.CountryCode = ValueOrDefault(order.DeliveryProfile.Country, "US");
	order.DeliveryProfile.Phone = ValueOrDefault(order.DeliveryProfile.Phone, "*");
	order.DeliveryProfile.Fax = ValueOrDefault(order.DeliveryProfile.Fax, "*");
	order.DeliveryProfile.Email = order.User.Username;

	-- Delivery Profile validation
	ValidateField(order.DeliveryProfile.FirstName, "Delivery Profile First Name", 1, 50, validationErrors);
	ValidateField(order.DeliveryProfile.LastName, "Delivery Profile Last Name", 1, 50, validationErrors);
	ValidateField(order.DeliveryProfile.CompanyName, "Delivery Profile Company Name (User Organization)", 1, 50, validationErrors);
	ValidateField(order.DeliveryProfile.Address1, "Delivery Profile Address1", 1, 50, validationErrors);
	ValidateField(order.DeliveryProfile.Address2, "Delivery Profile Address2", 1, 50, validationErrors);
	ValidateField(order.DeliveryProfile.City, "Delivery Profile City", 1, 50, validationErrors);
	ValidateField(order.DeliveryProfile.Zip, "Delivery Profile Zip", 1, 50, validationErrors);
	ValidateField(order.DeliveryProfile.CountryCode, "Delivery Profile Country Code", 2, 2, validationErrors);
	ValidateField(order.DeliveryProfile.Phone, "Delivery Profile Phone", 1, 50, validationErrors);
	ValidateField(order.DeliveryProfile.Email, "Delivery Profile E-mail", 1, 64, validationErrors);

	if (order.DeliveryProfile.CountryCode and order.DeliveryProfile.CountryCode == "US") then
		ValidateField(order.DeliveryProfile.StateCode, "Delivery Profile State Code", 2, 2, validationErrors);
	else
		ValidateField(order.DeliveryProfile.StateName, "Delivery Profile State Name", 1, 50, validationErrors);
	end

	-- Processing instructions
	order.ProcessingInstructions = {};
	order.ProcessingInstructions[1] = 1; -- Using default value per specification
	
	-- Customer References
	order.CustomerReferences = {};
	order.CustomerReferences[1] = TagProcessor.ReplaceTags("{TableField:Transaction.TransactionNumber}");

	return order, validationErrors;
end

function AddOrderNote(transactionNumber, order)
	log:Debug("[AddOrderNote]");

	local note = string.format("Reprints Desk order submitted (ID %s)\r\nhttp://www.reprintsdesk.com/landing/os.aspx?o=%s&r=%s", order.OrderID, order.OrderID, order.RndID);

	ExecuteCommand("AddNote", {transactionNumber, note});
end

function ValidateField(input, name, minLength, maxLength, errors)
	local inputString = nil;

	if (input ~= nil) then
		inputString = tostring(input);
	end

	if (inputString == nil or (minLength > 0 and #inputString == 0)) then
		table.insert(errors, name .. " - Required");
	elseif (inputString ~= nil and #inputString < minLength) then
		table.insert(errors, name .. " - Minimum length is " .. minLength .. " characters");
	elseif (inputString ~= nil and #inputString > maxLength) then
		table.insert(errors, name .. " - Maximum length is " .. maxLength .. " characters");
	end
end

function ValueOrDefault(value, defaultValue)
	local valueString = nil;

	if (value ~= nil) then
		valueString = tostring(value);
	end

	if (valueString ~= nil and (#valueString > 0)) then
		log:Debug("Using the value provided: " ..valueString);
		return value;
	else
		log:Debug("Using Default value");
		return defaultValue;
	end
end

function OnError(exception)
	local currentException = exception;

	while (currentException ~= nil) do
		log:Error(currentException.Message);		
		currentException = currentException.InnerException;
	end
end